import random
import math
import datetime
import numpy as np
import pandas as pd
import torch
from matplotlib import pyplot as plt
import seaborn as sns
from dl2050utils.core import *
from dl2050utils.ju import in_ipynb
from IPython.display import display
T = torch.tensor

def get_logprobs(model, x, cuda=True):
    x = torch.Tensor(x)
    if cuda and torch.cuda.is_available():
        x = x.cuda()
    with torch.no_grad():
        y2 = model(x)
    return y2.detach().cpu().numpy()

def get_logprobs_with_batch(model, x, bs=256, cuda=True):
    return np.vstack([get_logprobs(model, x[i*bs:(i+1)*bs], cuda=cuda) for i in range(math.ceil(len(x)/bs))])

def get_probs_from_logprobs(y2):
    return np.exp(np.max(y2, axis=1))
    return np.exp(y2[:,1])

def get_preds_from_logprobs(y2):
    return y2.argmax(axis=1)

def get_probs(model, x):
    y2 = get_logprobs(model, x)
    return get_probs_from_logprobs(y2)

def get_preds(model, x):
    y2 = get_logprobs(model, x)
    return get_preds_from_logprobs(y2)

def get_prob_positive(probs, preds):
    probs2 = np.zeros(len(probs))
    m = np.where(preds==1)
    probs2[m] = probs[m]
    return np.round(probs2,2)

def predict(model, x):
    x = torch.Tensor(x).cuda()
    with torch.no_grad():
        y2 = model(x)
    y2 = y2.detach().cpu().numpy()
    preds = y2.argmax(axis=1)
    return preds

def norm(x):
    m,s = x.mean(), x.std()
    return (x-m)/s if s>0 else x

def rebalance_xy(x, y, n=None, k=None):
    if n is None and k is None:
        return x,y
    k1 = len(np.where(y==1)[0])
    k0 = n if k is None else k1*k
    m = np.hstack((np.where(y==0)[0][:k0],np.where(y==1)[0]))
    return x[m],y[m]

def frmt_n(n): return f'{n:,}'.replace(',','.')

def show_clsf_stats(y, c, cls, c_szs, c_dist):
    df = pd.concat([
        pd.DataFrame(np.arange(0,len(cls)),columns=['cls']),
        pd.DataFrame(cls,columns=['labels']),
        pd.DataFrame(c_szs,columns=['size']),
        pd.DataFrame(c_dist,columns=['dist'])
    ], axis=1)
    print(f'\nClassification dataset: {frmt_n(len(y))} records, {frmt_n(c)} classes')
    if len(y.shape)>1: print(f'Total number of labels {y.shape} -> {np.asarray(c_szs).sum():,}')
    display(df)

def prep_cls(y, cls_to_idx=None, cls=None):
    assert (type(y)==list or type(y)==np.ndarray)
    
    size,labels,cls2idx,cls2 = 0,None,None,None
    if type(y)==list:
        size = len(y)
        labels = [str(e) for e in y]
        if cls_to_idx is not None:
            cls2idx = cls_to_idx
            cls2 = [e for e in cls2idx]
            for label in labels:
                if label not in cls2:
                    raise Exception(f'Label {label} not found in cls_to_idx')
        else:
            cls2 = sorted(list(set(labels)))
            cls2idx = {cls2[i]: i for i in range(len(cls2))}
        if type(y[0])==str:
            y = [cls2idx[label] for label in labels]
        y = np.asarray(y)
        c_idxs = [(y==cls2idx[c]).nonzero()[0] for c in cls2idx]
    else:
        y1 = y.flatten()
        size = len(y1)
        cls1 = np.unique(y1)
        c_idxs = [(y==cls1[c]).nonzero()[0] for c in cls1]
        cls2 = cls or cls1
    c_szs = [e.shape[0] for e in c_idxs]
    c_dist = [e.shape[0]/size for e in c_idxs]
    return y,cls2,len(cls2),labels,cls2idx,c_idxs,c_szs,c_dist

def np_split(x, y, p=.8):
    idx = np.random.permutation(x.shape[0])
    k = int(x.shape[0]*p)
    x,y = x[idx].copy(),y[idx].copy()
    return x[:k],y[:k],x[k:],y[k:]

def get_sample(x1, y1, x2, y2, sz=1000, sz2=None):
    if sz2 is None: sz2=sz
    p = np.random.permutation(len(x1))
    x1, y1 = x1[p], y1[p]
    p = np.random.permutation(len(x2))
    x2, y2 = x2[p], y2[p]
    return x1[:sz], y1[:sz], x2[:sz2], y2[:sz2]

def plot(x, y):
    plt.figure()
    plt.plot(x,y)
    plt.show()
    plt.pause(.001)

def imgrid(x, y):
    if len(x.shape) not in [3,4]: Exception('x dims must be 3 ou 4')
    if len(x.shape)==3: x = x[None,:,:,:]
    n = x.shape[0]
    nc = 4
    nr = math.ceil(n/nc)
    fig, axs = plt.subplots(nr, nc, figsize=(15,4*nr))
    if type(axs)!=np.ndarray: axs = [[axs]]
    elif len(axs.shape)==1: axs = [axs]
    for i in range(n):
        r,c = i//nc,i%nc
        ax = axs[r][c]
        xi = x[i] if x[i].shape[2]>2 else x[i].squeeze(-1)
        ax.imshow(xi)
        ax.set_title(y[i])
    fig.tight_layout()
    plt.show()
    plt.pause(0.01)

def plot_signal(x, ch=0, title=None):
    plt.plot(x[ch])
    if title: plt.title(title)
    plt.show()
    plt.pause(0.001)
    
def sigshow(x, title):
    n = x.shape[0]
    nc = 2
    nr = math.ceil(n/nc)
    fig, axs = plt.subplots(nr, nc, figsize=(15,2*nr))
    for i in range(n):
        r,c = i//nc,i%nc
        axs[r][c].plot(x[i])
        axs[r][c].set_title(f'Channel {i}')
    fig.tight_layout()
    plt.show()
    plt.pause(0.001)


# def np_split(x, y, p=.2):
#     mask = np.random.permutation(len(x))
#     k = int(len(x)*(1-p))
#     x1, y1 = x[mask[:k]], y[mask[:k]]
#     x2, y2 = x[mask[k:]], y[mask[k:]]
#     return x1, y1, x2, y2


# def get_sample(x1, y1, x2, y2, sz=1000):
#     p = np.random.permutation(len(x1))
#     x1, y1 = x1[p], y1[p]
#     p = np.random.permutation(len(x2))
#     x2, y2 = x2[p], y2[p]
#     return x1[:sz], y1[:sz], x2[:sz], y2[:sz]
# x1, y1, x2, y2 = get_sample(x1, y1, x2, y2)
# x1.shape, y1.shape, x2.shape, y2.shape


# torch.Tensor.ndim = property(lambda x: len(x.shape))
def _tuple(x): return x if isinstance(x, (list, tuple)) else (x,x)
NoneType = type(None)
def noop (x=None, *args, **kwargs): return x
def t2s(t): return f'{int(t)//60:02d}:{int(t)%60:02d}'


def get_date_str(): 
    d = datetime.datetime.now()
    return  f'{d.year:04d}{d.month:02d}{d.day:02d}-{d.hour:02d}{d.minute:02d}{d.second:02d}'

def get_log_fname(path):
    dirname = path/'runs'
    fname = f'{dirname}/run-{get_date_str()}.csv'
    return fname, dirname

def save_model(model, path, stats):
    state = {'state_dict': model.state_dict(), 'stats': stats}
    torch.save(state, path)

def load_model(model, path, cuda=False, inference=False, show=False):
    state = torch.load(path, map_location='cpu')
    model.load_state_dict(state['state_dict'])
    stats = state['stats']
    if show:
        pd.options.display.float_format = '{:,.4f}'.format
        display(pd.DataFrame.from_dict(stats))
    if cuda and torch.cuda.is_available():
        model.cuda()
    if inference:
        model.eval()
    return stats


def get_model_n_params(model):
    return sum([p.numel() for p in model.parameters()])

def xy_shuffle_split(x, y, p=0.2):
    msk = np.random.permutation(len(x))
    n = int(len(x)*p)
    return x[msk[:n]], y[msk[:n]], x[msk[n:]], y[msk[n:]]


def get_cls_dist(y, cls_to_idx):
    if type(y)==list: y = np.array([cls_to_idx[e] for e in y])
    elif type(y)==torch.Tensor: y = y.numpy()
    c_idxs = [np.asarray(y==cls_to_idx[c]).nonzero()[0] for c in cls_to_idx]
    c_szs = [e.shape[0] for e in c_idxs]
    c_dist = [e.shape[0]/len(y) for e in c_idxs]
    return c_idxs, c_szs, c_dist


def get_cls_dist_string(cls, c_szs, c_dist):
    return ', '.join([f'({c} {s} {d:.2f})' for c,s,d in sorted(zip(cls, c_szs, c_dist), key = lambda t: -t[1])])


def get_cnn_szs(net, sz):
    if sz is None: return []
    C0, H0, W0 = sz
    C1, H1, W1 = C0, H0, W0
    szs = [(C0, H0, W0)]
    for name, l in net.named_modules():
        classname = l.__class__.__name__
        if classname == 'Conv2d':
            stride = _tuple(l.stride)
            padding = _tuple(l.padding)
            kernel_size = _tuple(l.kernel_size)
            dilation = _tuple(l.dilation)
            C1 = int(l.out_channels)
            H1 = int(math.floor((H0 + 2 * padding[0] - dilation[0] * (kernel_size[0] - 1) - 1) / stride[0] + 1))
            W1 = int(math.floor((W0 + 2 * padding[1] - dilation[1] * (kernel_size[1] - 1) - 1) / stride[1] + 1))
            szs.append([(C1, H1, W1)])
        C0, H0, W0 = C1, H1, W1
    return szs


def biased_roll(dist):
    r = random.random()
    cum, result = 0, 0
    for d in dist:
        cum += d
        if r < cum: return result
        result += 1


def table(cols, vals):
    pd.options.display.float_format = '{:,.4f}'.format
    display(pd.DataFrame(vals, columns=cols))


# Deprecated
# def show_dist(y):
#     df = pd.DataFrame(y, columns=['count'])['count'].value_counts()
#     df = pd.DataFrame(df, index=df.index, columns=['count'])
#     df['p'] = df['count']/sum(df['count'])
#     display(df)
    # print(get_cls_dist_string(self.cls, self.c_szs, self.c_dist))


# def calc_norm(self):
#     xs, ys = [], []
#     for i in range(len(self)):
#         x, y = self[i]
#         xs.append(x)
#         ys.append(y)
#     x, y = torch.stack(xs, 0), torch.stack(ys, 0)
#     mean = [x[:,i,:,:].mean().item() for i in range(3)]
#     std = [x[:,i,:,:].std().item() for i in range(3)]
#     return mean, std



# # L()

# def is_iter(o):
#     "Test whether `o` can be used in a `for` loop"
#     #Rank 0 tensors in PyTorch are not really iterable
#     return isinstance(o, (Iterable,Generator)) and getattr(o,'ndim',1)


# def mask2idxs(mask):
#     "Convert bool mask or index list to index `L`"
#     if isinstance(mask,slice): return mask
#     mask = list(mask)
#     if len(mask)==0: return []
#     it = mask[0]
#     if hasattr(it,'item'): it = it.item()
#     if isinstance(it,(bool,NoneType,np.bool_)): return [i for i,m in enumerate(mask) if m]
#     #if isinstance(it,(bool,np.bool_)): return [i for i,m in enumerate(mask) if m]
#     return [int(i) for i in mask]


# def _is_array(x): return hasattr(x,'__array__') or hasattr(x,'iloc')


# def is_indexer(idx):
#     "Test whether `idx` will index a single item in a list"
#     return isinstance(idx,int) or not getattr(idx,'ndim',1)


# def _listify(o):
#     if o is None: return []
#     if isinstance(o, list): return o
#     if isinstance(o, str) or _is_array(o): return [o]
#     if is_iter(o): return list(o)
#     return [o]


# def coll_repr(c, max_n=10):
#     "String repr of up to `max_n` items of (possibly lazy) collection `c`"
#     return f'(#{len(c)}) [' + ','.join(itertools.islice(map(str,c), max_n)) + (
#         '...' if len(c)>10 else '') + ']'


# class CollBase:
#     "Base class for composing a list of `items`"
#     def __init__(self, items): self.items = items
#     def __len__(self): return len(self.items)
#     def __getitem__(self, k): return self.items[k]
#     def __setitem__(self, k, v): self.items[list(k) if isinstance(k,CollBase) else k] = v
#     def __delitem__(self, i): del(self.items[i])
#     def __repr__(self): return self.items.__repr__()
#     def __iter__(self): return self.items.__iter__()


# class L(CollBase):
#     "Behaves like a list of `items` but can also index with list of indices or masks"
#     _default='items'
#     def __init__(self, items=None, *rest, use_list=False, match=None):
#         if rest: items = (items,)+rest
#         if items is None: items = []
#         if (use_list is not None) or not _is_array(items):
#             items = list(items) if use_list else _listify(items)
#         if match is not None:
#             if is_coll(match): match = len(match)
#             if len(items)==1: items = items*match
#             else: assert len(items)==match, 'Match length mismatch'
#         super().__init__(items)

#     @property
#     def _xtra(self): return None
#     def _new(self, items, *args, **kwargs): return type(self)(items, *args, use_list=None, **kwargs)
#     def __getitem__(self, idx): return self._get(idx) if is_indexer(idx) else L(self._get(idx), use_list=None)
#     def copy(self): return self._new(self.items.copy())

#     def _get(self, i):
#         if is_indexer(i) or isinstance(i,slice): return getattr(self.items,'iloc',self.items)[i]
#         i = mask2idxs(i)
#         return (self.items.iloc[list(i)] if hasattr(self.items,'iloc')
#                 else self.items.__array__()[(i,)] if hasattr(self.items,'__array__')
#                 else [self.items[i_] for i_ in i])

#     def __setitem__(self, idx, o):
#         "Set `idx` (can be list of indices, or mask, or int) items to `o` (which is broadcast if not iterable)"
#         idx = idx if isinstance(idx,L) else _listify(idx)
#         if not is_iter(o): o = [o]*len(idx)
#         for i,o_ in zip(idx,o): self.items[i] = o_

#     def __iter__(self): return iter(self.items.itertuples() if hasattr(self.items,'iloc') else self.items)
#     def __contains__(self,b): return b in self.items
#     def __invert__(self): return self._new(not i for i in self)
#     def __eq__(self,b): return False if isinstance(b, (str,dict,set)) else all_equal(b,self)
#     def __repr__(self): return repr(self.items) if _is_array(self.items) else coll_repr(self)
#     def __mul__ (a,b): return a._new(a.items*b)
#     def __add__ (a,b): return a._new(a.items+_listify(b))
#     def __radd__(a,b): return a._new(b)+a
#     def __addi__(a,b):
#         a.items += list(b)
#         return a

#     def sorted(self, key=None, reverse=False):
#         if isinstance(key,str):   k=lambda o:getattr(o,key,0)
#         elif isinstance(key,int): k=itemgetter(key)
#         else: k=key
#         return self._new(sorted(self.items, key=k, reverse=reverse))

#     @classmethod
#     def split(cls, s, sep=None, maxsplit=-1): return cls(s.split(sep,maxsplit))

#     @classmethod
#     def range(cls, a, b=None, step=None):
#         if is_coll(a): a = len(a)
#         return cls(range(a,b,step) if step is not None else range(a,b) if b is not None else range(a))

#     def map(self, f, *args, **kwargs):
#         g = (bind(f,*args,**kwargs) if callable(f)
#              else f.format if isinstance(f,str)
#              else f.__getitem__)
#         return self._new(map(g, self))

#     def filter(self, f, negate=False, **kwargs):
#         if kwargs: f = partial(f,**kwargs)
#         if negate: f = negate_func(f)
#         return self._new(filter(f, self))

#     def unique(self): return L(dict.fromkeys(self).keys())
#     def enumerate(self): return L(enumerate(self))
#     def val2idx(self): return {v:k for k,v in self.enumerate()}
#     def itemgot(self, *idxs):
#         x = self
#         for idx in idxs: x = x.map(itemgetter(idx))
#         return x
    
#     def attrgot(self, k, default=None): return self.map(lambda o:getattr(o,k,default))
#     def cycle(self): return cycle(self)
#     def map_dict(self, f=noop, *args, **kwargs): return {k:f(k, *args,**kwargs) for k in self}
#     def starmap(self, f, *args, **kwargs): return self._new(itertools.starmap(partial(f,*args,**kwargs), self))
#     def zip(self, cycled=False): return self._new((zip_cycle if cycled else zip)(*self))
#     def zipwith(self, *rest, cycled=False): return self._new([self, *rest]).zip(cycled=cycled)
#     def map_zip(self, f, *args, cycled=False, **kwargs): return self.zip(cycled=cycled).starmap(f, *args, **kwargs)
#     def map_zipwith(self, f, *rest, cycled=False, **kwargs): return self.zipwith(*rest, cycled=cycled).starmap(f, **kwargs)
#     def concat(self): return self._new(itertools.chain.from_iterable(self.map(L)))
#     def shuffle(self):
#         it = copy(self.items)
#         random.shuffle(it)
#         return self._new(it)
    
#     def append(self,o): return self.items.append(o)
#     def remove(self,o): return self.items.remove(o)
#     def count (self,o): return self.items.count(o)
#     def reverse(self ): return self.items.reverse()
#     def pop(self,o=-1): return self.items.pop(o)
#     def clear(self   ): return self.items.clear()
#     def index(self, value, start=0, stop=sys.maxsize): return self.items.index(value, start=start, stop=stop)
#     def sort(self, key=None, reverse=False): return self.items.sort(key=key, reverse=reverse)





# class Normalize1():
#     def __init__(self, m, s, shift=[0,0,0]):
#         self.m, self.s, self.shift = torch.tensor(m), torch.tensor(s), torch.tensor(shift)
#     def __call__(self, x):
#         print(x.size())
#         print(self.m.size(), self.s.size(), self.shift.size())
#         return (x-self.m)/self.s - self.shift


import torch


def NLLLoss(y2, y):
    out = torch.zeros_like(y, dtype=torch.float)
    for i in range(len(y)):
        out[i] = y2[i][y[i]]
    return -out.sum()/len(out)


# learner.logger.loss_t
# learner.logger.n
# learner.logger.x.shape[0]
# learner.logger.loss_t/learner.logger.n


# y = learner.y.detach().cpu()
# y2 = learner.y2.detach().cpu()

# loss = NLLLoss()(y2, y)
# loss


import re

name = 'nn'
package = f'dl2050{name}'
description = 'NN lib'
author = 'João Neto'
author_email = 'joao.filipe.neto@gmail.com'
keywords = ['Neural Networks']

def get_version_parts(ver):
      res = re.split(r'\.', ver, maxsplit=2)
      if len(res)<3: raise RuntimeError('Unable to parse version number')
      return res

def get_minor(ver):
      res = get_version_parts(ver)
      return res[2]

def get_camel(ver):
      res = get_version_parts(ver)
      return f'{res[0]}_{res[1]}_{res[2]}'

def save_version(ver):
      with open(f'./{package}/__version__.py', 'w') as f: f.write(f'version = "{ver}"')

def pump_version(ver):
      res = get_version_parts(ver)
      ver2 = f'{res[0]}.{res[1]}.{str(int(res[2])+1)}'
      save_version(ver2)
      return ver2


import random
import numpy as np
import torch
import torchvision.transforms as transforms
from PIL import Image
from IPython.display import display
from dl2050utils.core import *
from dl2050utils.fs import get_dir_files, get_dir_dirs
from dl2050nn.etc import *
from dl2050nn.tabular import DataFrameProcessor

# if regr:
#     yrange = [df[c_dep].astype('float32').min(), df[c_dep].astype('float32').max()]

################################################################################################################################
#  DataLoader
#
# TODO: num_workers, pin_memory
################################################################################################################################

def get_balanced_idxs(c_idxs, p=0.):
    p = min(max(p,0.),1.)
    szs = np.array([len(e) for e in c_idxs])
    n = szs.sum()
    bp0 = np.array([len(c)/n for c in c_idxs])
    bp1 = np.array([1/len(c_idxs)]*len(c_idxs))
    bp = bp0 + p*(bp1-bp0)
    m = bp/bp0
    szs2 = np.round((szs*m/m.min())).astype(int)
    z = np.array([])
    for i in range(len(c_idxs)):
        for _ in range(szs2[i]//szs[i]):
            z = np.concatenate((z,np.array(c_idxs[i])))
        z = np.concatenate((z,np.array(c_idxs[i])[:szs2[i]%szs[i]]))
    z = np.random.permutation(z)
    return z.astype(int)

def collate(b):
    x,y = zip(*b)
    return torch.stack(x),torch.stack(y)

class SimpleSampler():
    def __init__(self, ds, bs):
        self.n,self.bs = len(ds),bs
    def set_balance(self, p):
        pass
    def __iter__(self):
        self.idxs = list(range(0,self.n))
        for i in range(0, self.n, self.bs): yield self.idxs[i:i+self.bs]

class RandomSampler():
    def __init__(self, ds, bs, c_idxs):
        self.n,self.bs,self.c_idxs,self.p = len(ds),bs,c_idxs,None
    def set_balance(self, p):
        self.p = p
    def __iter__(self):
        if self.p is None or self.c_idxs is None:
            self.idxs = list(np.random.permutation(np.arange(self.n)))
        else:
            self.idxs = list(get_balanced_idxs(self.c_idxs, p=self.p))
        for i in range(0, self.n, self.bs): yield self.idxs[i:i+self.bs]

class DataLoader():
    def __init__(self, ds, bs=64, shuffle=False, collate_fn=collate):
        self.ds, self.bs, self.collate_fn, self.n = ds, bs, collate_fn, len(ds)
        self.sampler = SimpleSampler(ds, bs) if not shuffle else RandomSampler(ds, bs, self.ds.c_idxs)
    def set_balance(self, p): self.sampler.set_balance(p)
    def __len__(self): return math.ceil(len(self.ds)/self.bs)
    def __iter__(self):
        for b in self.sampler: yield self.collate_fn([self.ds[i] for i in b])

################################################################################################################################
# Dataset, Data
#
# Classification: y type is list or numpy long
# Mask of integers: 2D classification
# Mask of floats: ???
# Regression: y type is float scalar
#
################################################################################################################################

class Dataset():
    def __init__(self, x, y, aug_size=None, clsf=False, cls_to_idx=None, cls=None, regr=False, trfs=[], show=True):
        self.x,self.y,self.aug_size,self.clsf,self.regr,self.trfs,self.show = x,y,aug_size,clsf,regr,trfs or [],show
        self.bps,self.c_idxs = None,None
        if clsf:
            y,self.cls,self.c,self.labels,self.cls_to_idx,self.c_idxs,self.c_szs,self.c_dist = prep_cls(y, cls_to_idx=cls_to_idx, cls=cls)
            self.y = y.astype(np.int32)
            if show: show_clsf_stats(self.y, self.c, self.cls, self.c_szs, self.c_dist)
        else:
            self.y = y.astype(np.float32)
        self.x = x #.astype(np.float64)
        
    def __len__(self): return self.aug_size or len(self.x)
    def __getitem__(self, i):
        if self.aug_size: i=random.randint(0,len(self.x)-1)
        x,y = self.load(self.x[i], self.y[i])
        for trf in self.trfs: x,y=trf(x,y)
        x = torch.from_numpy(x).float()
        y = torch.from_numpy(y) if len(y.shape)>0 else torch.tensor(y).float()
        if self.clsf: y = y.long()
        return x,y
    def load(self, x, y): return x,y
    def show_data(self, idxs, labels=None):
        for idx in listify(idxs):
            x,y = self[idx]
            print(x.shape, y.shape)
    def show_trfs(self, n=8): self.show([random.randint(0,len(self))]*n)
    def encode(self, x, y): pass
    def decode(self,x,y): return x.numpy(),y.item()


class Data:
    def __init__(self, ds1, ds2, bs=64, show=True):
        self.ds1,self.ds2, self.bs,self.show =  ds1,ds2,bs,show
        # w, pin = 0, torch.cuda.is_available()
        self.dl1 = DataLoader(ds1, bs=bs, shuffle=True)
        self.dl2 = DataLoader(ds2, bs=bs, shuffle=False) if ds2 is not None else None

    def show_batch(self, n=8): self.ds1.show(next(iter(self.dl1.sampler))[:n])
    def show_batch_2(self, n=8): self.ds2.show(next(iter(self.dl2.sampler))[:n])
    def show_trfs(self, n=8): self.ds1.show_trfs(n=n)

    @property
    def clsf(self): return self.ds1.clsf
    @property
    def regr(self): return self.ds1.regr
    @property
    def c(self): return self.ds1.c if self.clsf else None
    @property
    def cls(self): return self.ds1.cls if self.clsf else None
    @property
    def cls_to_idx(self): return self.ds1.cls_to_idx if self.clsf else None

    @classmethod
    def from_numpy(cls, x1, y1, x2, y2, aug_size1=None, aug_size2=None, clsf=False, regr=False, trfs1=[], trfs2=[], bs=64, show=True):
        ds1 = Dataset(x1, y1,aug_size=aug_size1, clsf=clsf, regr=regr, trfs=trfs1, show=show)
        ds2 = Dataset(x2, y2, aug_size=aug_size2, clsf=clsf, regr=regr, trfs=trfs2, show=show)
        return cls(ds1, ds2, bs=bs)

################################################################################################################################
#  SignalDataset, SignalData
################################################################################################################################

class SignalDataset(Dataset):    
    def __init__(self, *args, **kwargs):
        super(SignalDataset, self).__init__(*args, **kwargs)
    def show_data(self, idxs, labels=None):
        idxs, labels = listify(idxs), listify(labels)
        for i,idx in enumerate(idxs):
            x,y = self.decode(*self[idx])
            if self.clsf: y = self.cls[y]
            label = labels[i] if len(labels) else y
            sigshow(x, label or y)

class SignalData(Data):
    def __init__(self, x1, y1, x2, y2, clsf=False, regr=False, trfs1=[], trfs2=[], bs=64, show=True):
        ds1 = SignalDataset(x1, y1, clsf=clsf, regr=regr, trfs=trfs1, show=show)
        ds2 = SignalDataset(x2, y2, clsf=clsf, regr=regr, trfs=trfs2, show=show)
        super(SignalData, self).__init__(ds1, ds2, bs=bs)
        
    @classmethod
    def from_numpy(cls, x1, y1, x2, y2, clsf=False, regr=False, trfs1=[], trfs2=[], bs=64, show=True):
        # x1,x2 = torch.from_numpy(x1).float(),torch.from_numpy(x2).float()
        # if clsf: y1, y2 = list(y1), list(y2)
        # else: y1, y2 = torch.from_numpy(y1).float(), torch.from_numpy(y2).float()
        return cls(x1, y1, x2, y2, clsf=clsf, regr=regr, trfs1=trfs1, trfs2=trfs2, bs=bs, show=show)

################################################################################################################################
#  ImageDataset, ImageDatasetMem, ImageData
################################################################################################################################

class ImageDataset(Dataset):    
    def load(self,x,y): return Image.open(x),y
    def decode(self, x, y): return transforms.ToPILImage()(x),y
    def show_data(self, idxs, labels=None):
        idxs, labels = listify(idxs), listify(labels)
        x,y = zip(*[self.decode(*self[idx]) for idx in idxs])
        x = np.stack([np.array(x) for x in x])
        if self.clsf: y = [self.cls[e] for e in y]
        imgrid(x, labels or y)


class ImageDatasetMem(ImageDataset):
    def load(self,x,y): return x,y
    def decode(self,x,y): return x.numpy().transpose((1,2,0)),y


class ImageData(Data):
    def __init__(self, x1, y1, x2, y2, clsf=False, regr=False, trfs1=[], trfs2=[], bs=64, ds_class=ImageDataset, show=True):
        ds1 = ds_class(x1, y1, clsf=clsf, regr=regr, trfs=trfs1, show=show)
        ds2 = ds_class(x2, y2, clsf=clsf, regr=regr, trfs=trfs2, cls_to_idx=ds1.cls_to_idx, show=show)
        super(ImageData, self).__init__(ds1, ds2, clsf=clsf, regr=regr, bs=bs, show=show)
        
    @classmethod
    def from_folder(cls, path1, path2, types=None, clsf=False, regr=False, trfs1=[], trfs2=[], bs=64):
        x, y = [[],[]], [[],[]]
        cls_names = [d.stem for d in get_dir_dirs(path1)]
        for i,path in enumerate([path1, path2]):
            for c in cls_names:
                for f in get_dir_files(path/f'{c}', types):
                    x[i].append(f)
                    y[i].append(c)
        return cls(x[0], y[0], x[1], y[1], clsf=clsf, regr=regr, trfs1=trfs1, trfs2=trfs2, bs=bs)

    @classmethod
    def from_numpy(cls, x1, y1, x2, y2, clsf=False, regr=False, trfs1=[], trfs2=[], bs=64, show=True):
        x1, y1, x2, y2 = T(x1).float(), y1, T(x2).float(), y2
        return cls(x1, y1, x2, y2, clsf=clsf, regr=regr, trfs1=trfs1, trfs2=trfs2, bs=bs, ds_class=ImageDatasetMem, show=show)

################################################################################################################################
#   TabularDataset, TabularData
################################################################################################################################

class TabularDataset(Dataset):
    def __init__(self, df, dfp, proc, clsf=False, cls_to_idx=None, regr=False, show=True):
        self.df, self.dfp = df, dfp
        x  = dfp[proc.c2].values
        x = torch.from_numpy(x).float()
        y = [str(e) for e in df[proc.c_dep].values]
        super(TabularDataset, self).__init__(x, y, clsf=clsf, cls_to_idx=cls_to_idx, regr=regr)
        if show and len(proc.get_emb_szs()): print(f'Embeddings: {proc.get_emb_szs()}')

    def show_data(self, idxs, labels=None, cols=None):
        df = self.df.iloc[listify(idxs)][cols or self.df.columns]
        df['results'] = labels
        display(df)


class TabularData(Data):
    def __init__(self, df1, df1p, df2, df2p, proc, clsf=False, regr=False, bs=128):
        self.df1, self.df1p, self.df2, self.df2p, self.proc = df1, df1p, df2, df2p, proc
        ds1 = TabularDataset(df1, df1p, proc, clsf=clsf, regr=regr)
        ds2 = TabularDataset(df2, df2p, proc, clsf=clsf, regr=regr,  cls_to_idx=ds1.cls_to_idx) if df2p is not None else None
        super(TabularData, self).__init__(ds1, ds2, clsf=clsf, regr=regr, bs=bs)

    @classmethod
    def from_dfs(cls, df1, df2, c_cat, c_cont, c_dep, c_date=[], c_key=None, c_skip=[], maxhot=16, max_embs=4, embs={}, path='.', 
                    clsf=False, regr=False, name='tabular', bs=256):
        proc = DataFrameProcessor.from_df(df1, c_cat, c_cont, c_dep, c_date=c_date, c_key=c_key, c_skip=c_skip, maxhot=maxhot,
                                            max_embs=max_embs, embs=embs)
        df1p = proc(df1)
        proc.save_meta(f'{path}/{name}_proc.pickle')
        df1.reset_index(drop=True).to_feather(f'{path}/{name}_df1.feather')
        df1p.reset_index(drop=True).to_feather(f'{path}/{name}_df1p.feather')
        df2p = None
        if df2 is not None:
            df2p = proc(df2)
            df2.reset_index(drop=True).to_feather(f'{path}/{name}_df2.feather')
            df2p.reset_index(drop=True).to_feather(f'{path}/{name}_df2p.feather')
        return cls(df1, df1p, df2, df2p, proc, clsf=clsf, regr=regr, bs=bs)

    @classmethod
    def from_feather(cls, path='.', name='tabular', clsf=False, regr=False, bs=128, valid=True):
        proc = DataFrameProcessor.from_meta_file(f'{path}/{name}_proc.pickle')
        df1 = feather.read_dataframe(f'{path}/{name}_df1.feather')
        df1p = feather.read_dataframe(f'{path}/{name}_df1p.feather')
        if not valid: return cls(df1, df1p, None, None, proc, clsf=clsf, regr=regr, bs=bs)
        df2, df2p = None, None
        if Path(f'{path}/{name}_df2.feather').is_file():
            df2 = feather.read_dataframe(f'{path}/{name}_df2.feather')
            df2p = feather.read_dataframe(f'{path}/{name}_df2p.feather')
        return cls(df1, df1p, df2, df2p, proc, clsf=clsf, regr=regr, bs=bs)

    def encode(self, df): return torch.from_numpy(self.proc(df)[self.proc.c2].values).float()
    def decode(self): pass

    @property
    def ncats(self): return self.proc.ncats
    @property
    def nconts(self): return self.proc.nconts
    @property
    def maxhot(self): return self.proc.maxhot
    @property
    def cat_szs(self): return self.proc.cat_szs
    @property
    def emb_szs(self): return self.proc.emb_szs



import math
from functools import partial
import torch
from torch import tensor
from dl2050utils.core import listify
import matplotlib.pyplot as plt
from dl2050nn.etc import *
from dl2050nn.cbs import *


def merge(os, dest, f):
    for o in os:
        for k,v in f(o).items():
            if k not in dest: dest[k] = v

def get_defaults(d): return getattr(d, '_defaults', {})


def get_p(epf, eps, cycle):
    p = math.floor(epf)/eps if cycle is None or cycle==0 else epf/cycle - math.floor(epf/cycle)
    if p>1.: p = p-math.floor(p)
    return p


def annealer(f):
    def _inner(start, end, cycle): return partial(f, start, end, cycle)
    return _inner

@annealer
def sched_no(start, end, cycle, epf, eps):  return start
@annealer
def sched_lin(start, end, cycle, epf, eps): return start+get_p(epf, eps, cycle)*(end-start)
@annealer
def sched_exp(start, end, cycle, epf, eps): return start*(end/start) ** get_p(epf, eps, cycle)
@annealer
def sched_cos(start, end, cycle, epf, eps): return start+(1+math.cos(math.pi*(1-get_p(epf, eps, cycle))))*(end-start)/2


def combine_scheds(pcts, scheds, cycle):
    assert sum(pcts) == 1.
    pcts = tensor([0] + listify(pcts))
    assert torch.all(pcts >= 0)
    pcts = torch.cumsum(pcts, 0)
    def _inner(epf, eps):
        pos = get_p(epf, eps, cycle)
        idx = (pos>=pcts).nonzero(as_tuple=False).max()
        if idx == len(pcts)-1: idx -= 1
        actual_pos = (pos-pcts[idx]) / (pcts[idx+1]-pcts[idx])
        return scheds[idx](actual_pos*cycle, eps)
    return _inner


def combine_scheds_mult(f1, f2):
    def _inner(epf, eps):
        return f1(epf, eps)*f2(epf, eps)
    return _inner


def sched_cycle(start, end, cycle=1., sched2=None, factor=.1):
    f1 = combine_scheds([0.3, 0.7], [sched_cos(start, end, cycle), sched_cos(end, start, cycle)], cycle)
    if sched2 is None: return f1
    return combine_scheds_mult(f1, sched2(1., factor, None))


class ParamScheduler(Callback):
    def __init__(self, pname, sched_funcs):
        self.pname,self.sched_funcs = pname,listify(sched_funcs)

    def plot(self, eps=4, label=None):
        fig = plt.figure(figsize=(5, 3))
        epf = torch.linspace(0.0,eps,100*eps)
        plt.plot(epf, [self.sched_funcs[0](e, eps) for e in epf], label=label)
        plt.show()
        plt.pause(.001)

    def begin_batch(self): 
        if not self.train: return
        # Pytorch optimizer
        if not hasattr(self.optf, 'hypers'):
            for pg in self.optf.param_groups:
                pg['lr'] = self.sched_funcs[0](self.epf, self.eps)
            return
        # NN2 optimizer
        fs = self.sched_funcs
        if len(fs)==1: fs = fs*len(self.optf.param_groups)
        for f,h in zip(fs,self.optf.hypers): h[self.pname] = f(self.epf, self.eps)


class Optimizer():
    def __init__(self, params, steppers=[], stats=None, **defaults):
        params = L(params)
        self.param_groups = L(L(p) for p in params) if isinstance(params[0], (L,list)) else L([params])
        self.steppers, self.stats = L(steppers), L(stats)
        merge(self.steppers, defaults, get_defaults)
        self.hypers = [{**defaults} for _ in self.param_groups]
        self.state = {}
        for p,_ in self.all_params():
            self.state[p] = {}
        
    def all_params(self):
        return [(p,hyper) for pg,hyper in zip(self.param_groups,self.hypers)
            for p in pg if p.grad is not None]
        
    def step(self):
        for p,hyper in self.all_params():
            for stat in self.stats:
                if p not in self.state: self.state[p] = {}
                self.state[p] = stat(self.state[p], p, **hyper)
            for step in self.steppers: step(p, **{**self.state[p], **hyper})

    def zero_grad(self):
        for p,_ in self.all_params():
            p.grad.detach_()
            p.grad.zero_()


# Stats

def average_grad(state, p, mom, dampening=False, **kwargs):
    "Keeps track of the avg grads of `p` in `state` with `mom`."
    if 'grad_avg' not in state: state['grad_avg'] = torch.zeros_like(p.grad.data)
    damp = 1-mom if dampening else 1.
    # state['grad_avg'].mul_(mom).add_(damp, p.grad.data) -> OLD DEPRECRATED
    state['grad_avg'].mul_(mom).add_(p.grad.data, alpha=damp) # NEW TO BE TESTED
    return state
average_grad.defaults = dict(mom=0.9)


def average_sqr_grad(state, p, sqr_mom, dampening=True, **kwargs):
    if 'sqr_avg' not in state: state['sqr_avg'] = torch.zeros_like(p.grad.data)
    damp = 1-sqr_mom if dampening else 1.
    # state['sqr_avg'].mul_(sqr_mom).addcmul_(damp, p.grad.data, p.grad.data) # -> OLD DEPRECRATED
    # state['sqr_avg'].mul_(sqr_mom).addcmul_(p.grad.data, p.grad.data, damp) # NEW TO BE TESTED
    state['sqr_avg'].mul_(sqr_mom).addcmul_(p.grad.data, p.grad.data, value=damp)
    return state
average_sqr_grad.defaults = dict(sqr_mom=0.99)


def step_stat(state, p, **kwargs):
    if 'step' not in state: state['step'] = 0
    state['step'] += 1
    return state


# Steps

def weight_decay(p, lr, wd, do_wd=True, **kwargs):
    if do_wd and wd!=0: p.data.mul_(1 - lr*wd)
    return p
#weight_decay.defaults = dict(wd=0.)


def l2_reg(p, lr, wd, do_wd=True, **kwargs):
    if do_wd and wd!=0:
        p.grad.data.add_(p.data, alpha=wd)
    return p
#l2_reg.defaults = dict(wd=0.)


def sgd_step(p, lr, **kwargs):
    p.data.add_(p.grad.data, alpha=-lr)
    return p
#sgd_step._defaults = dict(lr=0.01)


def momentum_step(p, lr, grad_avg, **kwargs):
    p.data.add_(grad_avg, alpha=-lr)
    return p


def debias(mom, damp, step): return damp * (1 - mom**step) / (1-mom)


def adam_step(p, lr, mom, step, sqr_mom, grad_avg, sqr_avg, epsilon, **kwargs):
    "Step for Adam with `lr` on `p`"
    debias1 = debias(mom, 1-mom, step)
    debias2 = debias(sqr_mom, 1-sqr_mom, step)
    # p.data.addcdiv_(-lr / debias1, grad_avg, (sqr_avg/debias2).sqrt() + epsilon)
    p.data.addcdiv_(grad_avg, (sqr_avg/debias2).sqrt()+epsilon, value=-lr/debias1)
    return p
#adam_step._defaults = dict(epsilon=1e-5, lr=0.01,)


def lamb_step(p, lr, mom, step, sqr_mom, grad_avg, sqr_avg, epsilon, **kwargs):
    "Step for LAMB with `lr` on `p`"
    debias1 = debias(mom, 1-mom, step)
    debias2 = debias(sqr_mom, 1-sqr_mom, step)
    r1 = p.data.pow(2).mean().sqrt()
    step = (grad_avg/debias1) / ((sqr_avg/debias2).sqrt()+epsilon)
    r2 = step.pow(2).mean().sqrt()
    q = 1 if r1 == 0 or r2 == 0 else min(r1/r2,10)
    p.data.add_(step, alpha=-lr*q)
    return p
#lamb_step._defaults = dict(epsilon=1e-6, wd=0.)


# Optimizers

def SGD(params, lr=0.01, mom=0., wd=1e-5, decouple_wd=True):
    steppers = [weight_decay] if decouple_wd else [l2_reg]
    steppers.append(sgd_step if mom==0 else momentum_step)
    if mom == 0.:
        return Optimizer(params, steppers, lr=lr, wd=wd)
    else:
        return Optimizer(params, steppers, stats=average_grad, lr=lr, mom=mom, wd=wd)
    
    
def Adam(params, lr=0.01, mom=0.9, sqr_mom=0.99, epsilon=1e-5, wd=1e-5):
    steppers = [l2_reg, adam_step]
    stats = [partial(average_grad, dampening=True), average_sqr_grad, step_stat]
    return Optimizer(params, steppers=steppers, stats=stats, lr=lr, mom=mom, sqr_mom=sqr_mom, epsilon=epsilon, wd=wd)


def Lamb(params, lr=0.01, mom=0.9, sqr_mom=0.99, epsilon=1e-5, wd=1e-5):
    steppers = [l2_reg, lamb_step]
    stats = [partial(average_grad, dampening=True), average_sqr_grad, step_stat]
    return Optimizer(params, steppers=steppers, stats=stats, lr=lr, mom=mom, sqr_mom=sqr_mom, epsilon=epsilon, wd=wd)


import torch
from functools import partial
import shutil
from dl2050nn.cbs import CancelBatchException, CancelEpochException, CancelTrainException, LR_Find, Recorder
from dl2050nn.optimizers import *
from dl2050nn.log import LoggerCallback
from dl2050nn.etc import *

class Learner():

    def __init__(self, data, model, opt=Adam, loss_f=None, metrics=[], cbs=[], device=None, model_name=None, pre=None, path='/data', save=True, show=True):
        if not path.is_dir(): raise Exception()
        if not (path/'models').exists(): (path/'models').mkdir()
        self.model = model
        self.model_name,self.path = model_name or str(model.__class__.__name__),path
        self.data = data
        self.device = 'cpu' if not torch.cuda.is_available() else torch.device(0) if device is None else torch.device(device)
        if show: print(f'Learner: device={self.device}\nBatch size: {data.bs}')
        self.opt = opt
        self.optf = self.opt(model.parameters())
        self.loss_f = loss_f
        if self.loss_f is not None: self.loss_f = self.loss_f.to(self.device)
        self.metrics = metrics
        self._init_cbs([LoggerCallback]+cbs)
        self.ep, self.train = 0, False
        self._save,self.show = save,show
        if pre:
            self.load(model_name=pre, update_best=False, show=show)
        self.model.to(self.device)

    @property
    def df_best(self): return self.logger.df_best

    def load(self, model_name=None, update_best=True, show=True):
        p = self.path/'models'/f'{model_name or self.model_name}.pth.tar'
        stats = load_model(self.model, p, show=show)
        if not update_best: return
        self.eval()
        self.logger.df_best = pd.DataFrame.from_dict(stats)
        k = 4+len(self.metrics)
        self.logger.best = self.logger.df_best.iloc[0,k]

    def save(self, model_name=None):
        p = self.path/'models'/f'{model_name or self.model_name}.pth.tar'
        save_model(self.model, p, self.logger.df_best.to_dict())

    def backup_model(self, backup_name):
        p = self.path/f'models/{backup_name}.pth.tar'
        shutil.copy(self.path/f'models/{self.model_name}.pth.tar', p)
        print(f'Model backup up: {p}')
    
    def _init_cbs(self, cbs):
        self.cbs = []
        for cb in cbs:
            cb = cb()
            cb.set_learner(self)
            setattr(self, cb.name, cb)
            self.cbs.append(cb)

    def _remove_cbs(self):
        for cb in self.cbs:
            cb.learner = None
        if hasattr(self, cb.name): delattr(self, cb.name)
        self.cbs.remove(cb)

    def __call__(self, event):
        res = True
        # for cb in sorted(self.cbs, key=lambda x: x._order):
        for cb in self.cbs: res = cb(event) and res
        return res

    def fit(self, eps):
        self.eps, self.loss = eps, torch.tensor(0.)
        try:
            if self('begin_fit'): return
            for ep in range(eps):
                self.ep = ep
                self.epf = ep
                self.train = True
                self.model.train()
                if not self('begin_epoch'): self.all_batches(self.data.dl1)
                self.train = False
                self.model.eval()
                with torch.no_grad(): 
                    if not self('begin_validate'): self.all_batches(self.data.dl2)
                if self('after_epoch'): break
        except CancelTrainException: self('after_cancel_train')
        finally:
            self('after_fit')

    def all_batches(self, dl):
        if not dl: return
        self.iters = len(dl)
        try:
            for self.iter,(x,y) in enumerate(dl):
                if len(x)>2: self.one_batch(x, y)
                # else: print('Skiping batch smaller that 2')
                if self.train: self.epf += 1./self.iters
        except CancelEpochException: self('after_cancel_epoch')

    def one_batch(self, x, y):
        try:
            self.x, self.y = x, y
            if self('begin_batch'): return
            self.x, self.y = x.to(self.device), y.to(self.device)
            self.y2 = self.model(self.x)
            self.y2 = torch.squeeze(self.y2, -1)
            if self('after_pred'): return
            self.loss = self.loss_f(self.y2, self.y)
            if self('after_loss'): return
            if not self.train: return
            self.loss.backward()
            if self('after_backward'): return
            self.optf.step()
            if self('after_step'): return
            self.optf.zero_grad()
        except CancelBatchException: self('after_cancel_batch')
        finally: self('after_batch')

    def lr_find(self, max_iter=100, min_lr=1e-6, max_lr=1e1):
        # Backup all cbs
        self._init_cbs([partial(LR_Find, max_iter=max_iter, min_lr=min_lr, max_lr=max_lr), Recorder])
        self.fit(10)
        # Reset optimizer status

    def eval(self):
        if not self.data.dl2:
            return
        self.model.eval()
        with torch.no_grad(): 
            if not self('begin_validate'): self.all_batches(self.data.dl2)

    def predict(self, x):
        x = self.data.encode(x)
        self.model.eval()
        with torch.no_grad():
            x = x.to(self.device)
            y = self.model(x)
        return y.detach().cpu()

    def close(self):
        self._remove_cbs()



import time
import os
from pathlib import Path
import numpy as np
import pandas as pd
import pickle
import gc
import torch
import matplotlib.pyplot as plt
from IPython.display import display
from dl2050utils.ju import disablePrint, enablePrint, force_print
from dl2050nn.etc import *
from dl2050nn.cbs import Callback
from dl2050nn.metrics import *

def key_metric(learner): return 4+len(learner.metrics)

def metrics_str(metrics):
    if not len(metrics): return ''
    s = ' M: ['
    for v in metrics: s += f'{v:.4f}, '
    s = s[:-2] + ']'
    return s


def log_str(ep, t, loss1, loss2, metrics1, metrics2, new_best, valid=False):
    s = f'Ep: {ep}\tT: {t2s(t)}:{int(t)%60:02d}\tL1: {loss1:.4f} '
    s += metrics_str(metrics1)
    if valid:
        s += f'\tL2: {loss2:.4f} '
        s += metrics_str(metrics2)
    if new_best: s += ' *'
    return s


def printProgressBar(i, n, prefix, w=50):
    p = float(i+1)/n
    filled = int(w*p)
    s = f'\033[31m{"█"*filled}\033[37m{"█"*(w-filled)}\033[0m'
    if p<1.: print(f'\r{prefix} | {s} | {100*p:.1f}%', end='\r')
    else: print(f'\r{" "*(w+len(prefix)+25)}', end='\r')


def plot_init(x_min, x_max):
    plt.ion()
    fig = plt.figure(figsize=(12, 3))
    ax = fig.add_axes([0.2, 0.2, .8, 0.8])
    ax.xaxis.set_ticks(np.arange(x_min, x_max, max(1,int((x_max-x_min))//10)))
    ax.yaxis.set_ticks(np.arange(0., 2.5, .5))
    ax.plot([], [], [], [],)
    return fig, ax


def plot_update(fig, ax, x, y1, y2):
    ax.lines[0].set_xdata(x), ax.lines[0].set_ydata(y1)
    if y2 is not None: ax.lines[1].set_xdata(x), ax.lines[1].set_ydata(y2)
    ax.relim()
    ax.autoscale()
    fig.canvas.draw()


def plot_df(df):
    fig = plt.figure(figsize=(12, 3))
    ax1, ax2 = fig.add_subplot(121), fig.add_subplot(122)
    df.plot(x='Epoch', y=['Loss_1', 'Loss_2'], ax=ax1)
    df.plot(x='Epoch', y=['accuracy_1', 'accuracy_2'], ax=ax2)
    plt.show()
    time.sleep(.01)


class LoggerCallback(Callback):
    
    def __init__(self, mode=None, file_=False, maximize=True, ep_stats=True):
        self.file, self.maximize, self.ep_stats = file_, maximize, ep_stats
        self.mode = ('ipynb' if in_ipynb() else 'console') if mode is None else mode
        if file_:
            fname, dirname = get_log_fname(self.path)
            if not os.path.exists(dirname): os.makedirs(dirname)
            self.logf = open(fname, 'w', 1)
        self.best = 0.0 if self.maximize else 1e9
        self.results = []
        self.y_ep, self.y2_ep, self.stats_1, self.stats_2, self.t0, self.t1 = None, None, None, None, 0, 0
        self.df, self.df_best = None, None
        pd.options.display.float_format = '{:,.4f}'.format
        
    def reset_stats(self):
        self.y_ep = torch.tensor([]).long() if self.data.clsf else torch.tensor([]).float()
        self.y2_ep = torch.tensor([]).float()
        self.loss_t, self.n, self.metrics_v = 0., 0, [0.]*len(self.metrics)
        
    def stats(self):
        if self.y_ep is None or self.y2_ep is None: return [0.]*(1+len(self.metrics))
        self.metrics_v = [m(self.y2_ep, self.y_ep) for m in self.metrics]
        return [self.loss_t/self.n] + self.metrics_v
    
    def accumulate(self):
        k = self.x.shape[0]
        loss = self.loss.detach().cpu().item()
        self.loss_t += loss*k
        self.n += k
        self.y_ep = torch.cat((self.y_ep, self.y.detach().cpu()))
        self.y2_ep = torch.cat((self.y2_ep, self.y2.detach().cpu()))

    def update_historic(self):
        row = [self.ep, self.t2] + [self.loss1] + [e for e in self.metrics1] + [self.loss2] + [e for e in self.metrics2]
        self.results.append(row)
        self.df = pd.DataFrame(self.results, columns=self.cols)
       
    def log_epoch(self):
        if self.mode == 'ipynb' and self.show and self.eps > 1:
            x = self.df['Epoch'].values
            y = self.df[['Loss_1', 'Loss_2']].values
            y0, y1 = y[:,0], y[:,1] if self.data.dl2 is not None else None
            plot_update(self.fig, self.ax, x, y0, y1)
        s = log_str(self.ep, self.t2, self.loss1, self.loss2, self.metrics1, self.metrics2, self.new_best, valid=self.data.dl2 is not None)
        if self.show: print(s)

    def begin_fit(self):
        if self.data.dl2 is None: self.best = 1e9
        if self.mode == 'ipynb' and self.show and self.eps > 1: self.fig, self.ax = plot_init(0, self.eps)
        self.cols = ['Epoch', 'Time', 'Loss_1'] + [f'{e.__name__}_1' for e in self.metrics] + ['Loss_2'] + [f'{e.__name__}_2' for e in self.metrics]
        if self.show:
            if self.file:
                print(f'Logging to file: {self.fname}')
            else:
                print(f'Logging to {self.mode}')
        
    def begin_epoch(self):
        self.t0 = time.time()
        self.reset_stats()
        
    def begin_validate(self):
        self.t1 = time.time() - self.t0
        self.stats_1 = [v for v in self.stats()]
        self.reset_stats()
            
    def after_loss(self):
        with torch.no_grad(): self.accumulate()
        
    def after_batch(self):
        if self.show: printProgressBar(self.iter, self.iters, f'Ep: {self.ep:d} {"train" if self.train else "eval"}:')

    def after_epoch(self):
        self.t2 = time.time() - self.t0 if self.t0 is not None else time.time()
        self.stats_2 = [v for v in self.stats()] if self.data.dl2 is not None else [0]*len(self.stats_1)
        self.loss1, self.loss2 = self.stats_1[0], self.stats_2[0] if self.data.dl2 is not None else 0.
        self.metrics1, self.metrics2 = [v for v in self.stats_1[1:]], [v for v in self.stats_2[1:]]
        loss_metric = not len(self.metrics) or self.data.dl2 is None
        self.new_best = False
        if loss_metric:
            if self.data.dl2 is None:
                self.new_best = self.loss1 < self.best
                if self.new_best: self.best = self.loss1
            else:
                self.new_best = self.loss2 < self.best
                if self.new_best: self.best = self.loss2
        else:
            self.new_best = (self.maximize and self.metrics2[0]>self.best) or (not self.maximize and self.metrics2[0]<self.best)
            if self.new_best: self.best = self.metrics2[0]
        self.update_historic()
        self.log_epoch()
        if self.new_best:
            self.df_best = self.df[self.df['Epoch']==self.ep]
            if self._save: self.save()
        
    def after_fit(self):
        if self.show: print(f'\nBest: {self.best:.4f}')
        plt.ioff()

    def show_best(self): display(self.df_best)

    def plot_loss1_metric2(self):
        if self.df is None: return
        col_ids = [2, 4+len(self.metrics)]
        z = self.df.iloc[:,col_ids].values
        plot(*list(zip(*z))); plt.show(); time.sleep(.01)



from pathlib import Path
import pandas as pd
import pickle
import gc
from IPython.display import display
from dl2050nn.etc import load_model, get_logprobs
from dl2050nn.results import Results_Clsf
from dl2050nn.log import key_metric

class Runner:
    def __init__(self, path, model=None, data=None, get_learner=None, nruns=10, eps=10, name='Runner', desc='', pre=None):
        self.path,self.model = path,model
        self.get_learner,self.data = get_learner,data
        self.nruns,self.eps,self.name,self.desc,self.pre = nruns,eps,name,desc,pre
        if not Path(f'{path}/runners').exists(): Path(f'{path}/runners').mkdir()
        self.fname = f'{path}/runners/{name}'
        self.learner,self.df,self.best = None,None,0.
            
    def __call__(self, show=False):
        self.model = self.model.cuda()
        self.best, self.best_file = 0., None
        for i in range(self.nruns):
            if show: print(f'\rRun {i}')
            learner = self.get_learner(model_name= f'{self.name}_{i}', pre=self.pre)
            self.learner = learner
            learner.fit(self.eps)
            learner.load(show=show)
            df1 = learner.logger.df_best
            df1['Best'] = '*' if learner.logger.best > self.best else ''
            self.df = df1.copy() if self.df is None else pd.concat((self.df,df1),ignore_index=True)
            if learner.logger.best > self.best:
                self.best,self.best_file = learner.logger.best,f'{self.name}_{i}'
                display(self.df.iloc[[-1]])
            self.save()
            learner.close()
            del learner
            gc.collect()
        self.save()
        print(f'\nBest file: {self.best_file}')
        if show: display(self.df)

    def save(self):
        meta = {'desc': self.desc, 'best': self.best, 'best_file': self.best_file}
        pickle.dump(meta, open(self.fname+'.pickle', 'wb'))
        self.df.to_csv(self.fname+'.csv', index=False)

    def load(self, show=False):
        p_pickle,p_csv = Path(f'{self.fname}.pickle'),Path(f'{self.fname}.csv')
        if not p_pickle.exists() or not p_csv.exists(): return
        meta = pickle.load(open(self.fname+'.pickle', 'rb'))
        if show: print('cols' in meta, 'best' in meta,  'best_file' in meta)
        if not 'best' in meta or not 'best_file' in meta: return
        if show: print(f'Loaded from file {self.fname}: {meta}')
        self.desc, self.best, self.best_file = meta['desc'], meta['best'], meta['best_file']
        self.df = pd.read_csv(self.fname+'.csv')
        if show: display(self.df)
        self.df['Best'] = self.df['Best'].fillna('')
        load_model(self.model, f'{self.path}/models/{self.best_file}.pth.tar', show=show, cuda=True, inference=True)

    def df_best(self):
        if self.learner is None: return None
        return self.df[self.df.iloc[:,key_metric(self.learner)]==self.best].iloc[[0]]
    
    def stats(self):
        return dict(self.df[self.df['accuracy_2']==self.df['accuracy_2'].max()].iloc[0])
    
    def show_results(self):
        self.load()
        y2,y = get_logprobs(self.model, self.data.ds2.x),self.data.ds2.y
        r = Results_Clsf(y2=y2, y=y)
        r.confusion()
        r.roc()
        r.stats()
        return r
